package com.huaching.parking.local.parking_local_general.config.zhaf;

import org.apache.http.HttpEntityEnclosingRequest;
import org.apache.http.HttpRequest;
import org.apache.http.NoHttpResponseException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.HttpRequestRetryHandler;
import org.apache.http.client.methods.HttpRequestWrapper;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.conn.ConnectTimeoutException;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.protocol.HttpContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.client.ClientHttpRequestFactory;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.web.client.RestTemplate;

import javax.net.ssl.SSLException;
import javax.net.ssl.SSLHandshakeException;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.UnknownHostException;

/**
 * @program: ths-scheduller
 * @description: RestTemplate连接池调用
 * @author: lidongsheng
 * @createData: 2019-12-09 17:02
 * @updateAuthor: lidongsheng
 * @updateData: 2019-12-09 17:02
 * @updateContent:
 * @Version: 1.0.0
 */
@Configuration
public class RestTemplatePoolConfig {

    @Autowired
    private RestTemplateBuilder builder;

    private static final Logger logger = LoggerFactory.getLogger(RestTemplatePoolConfig.class);

    @Bean
    public RestTemplate restTemplate() {
        RestTemplate restTemplate = builder.build();
        restTemplate.setRequestFactory(clientHttpRequestFactory(3000, 3000, 10000));
        return restTemplate;
    }

    /**
     * 链接线程池管理,可以keep-alive不断开链接请求,这样速度会更快
     * MaxTotal 连接池最大连接数
     * DefaultMaxPerRoute 每个主机的并发
     * ValidateAfterInactivity 可用空闲连接过期时间,重用空闲连接时会先检查是否空闲时间超过这个时间，如果超过，释放socket重新建立
     */
    private HttpClient getHttpClient() {
        PoolingHttpClientConnectionManager poolingConnectionManager = new PoolingHttpClientConnectionManager();
        poolingConnectionManager.setMaxTotal(1000);
        poolingConnectionManager.setDefaultMaxPerRoute(500);
        poolingConnectionManager.setValidateAfterInactivity(30000);
        /**
         * 设置HTTP连接管理器,连接池相关配置管理
         */
        HttpClientBuilder httpClientBuilder = HttpClientBuilder.create();
        httpClientBuilder.setConnectionManager(poolingConnectionManager);
        httpClientBuilder.setRetryHandler((e, i, httpContext) -> {
            HttpRequestWrapper httpRequestWrapper = (HttpRequestWrapper) httpContext.getAttribute("http.request");
            if (i > 2) {
                logger.error("【连接重试】 超过2次 放弃请求:" + httpRequestWrapper.getOriginal().toString());
                return false;
            }
            logger.error("【连接重试】 第 {} 次重试请求 -> {}:", i, httpRequestWrapper.getOriginal().toString());
            if (e instanceof NoHttpResponseException) {
                //服务器没有响应,可能是服务器断开了连接,应该重试
                logger.error("【连接重试】 服务器没有响应 重试:" + httpRequestWrapper.getOriginal().toString());
                return true;
            }
            if (e instanceof SSLHandshakeException) {
                // SSL握手异常
                logger.error("【连接重试】 SSL握手异常 重试:" + httpRequestWrapper.getOriginal().toString());
                return false;
            }
            if (e instanceof InterruptedIOException) {
                //超时
                logger.error("【连接重试】 超时 重试:" + httpRequestWrapper.getOriginal().toString());
                return true;
            }
            if (e instanceof UnknownHostException) {
                // 服务器不可达
                logger.error("【连接重试】 服务器不可达 不重试:" + httpRequestWrapper.getOriginal().toString());
                return false;
            }
            if (e instanceof ConnectTimeoutException) {
                // 连接超时
                logger.error("【连接重试】 连接超时 重试:" + httpRequestWrapper.getOriginal().toString());
                return true;
            }
            if (e instanceof SSLException) {
                logger.error("【连接重试】 SSLException 不重试:" + httpRequestWrapper.getOriginal().toString());
                return false;
            }
            logger.error("未知IOException异常:" + e.getMessage());
            HttpClientContext context = HttpClientContext.adapt(httpContext);
            HttpRequest request = context.getRequest();
            if (!(request instanceof HttpEntityEnclosingRequest)) {
                logger.error("【连接重试】 请求不是关闭连接的请求 重试:" + httpRequestWrapper.getOriginal().toString());
                //如果请求不是关闭连接的请求
                return true;
            }
            logger.error("【连接重试】 不重试:" + httpRequestWrapper.getOriginal().toString());
            return false;
        });
        return httpClientBuilder.build();
    }


    /**
     * 客户端请求连接策略
     *
     * @param ConnectTimeout           指客户端和服务器建立连接的timeout，
     * @param ReadTimeout              指客户端从服务器读取数据的timeout
     * @param ConnectionRequestTimeout 指从连接池获取连接的timeout
     * @return
     */
    private ClientHttpRequestFactory clientHttpRequestFactory(int ConnectTimeout, int ReadTimeout, int ConnectionRequestTimeout) {
        HttpComponentsClientHttpRequestFactory clientHttpRequestFactory = new HttpComponentsClientHttpRequestFactory();
        clientHttpRequestFactory.setHttpClient(getHttpClient());
        // 连接超时时间/毫秒
        clientHttpRequestFactory.setConnectTimeout(ConnectTimeout);
        // 读写超时时间/毫秒
        clientHttpRequestFactory.setReadTimeout(ReadTimeout);
        // 请求超时时间/毫秒
        clientHttpRequestFactory.setConnectionRequestTimeout(ConnectionRequestTimeout);
        return clientHttpRequestFactory;
    }
}
