package com.huaching.parking.local.parking_local_general.domain.local.bo;

import com.alibaba.fastjson.JSON;
import com.huaching.parking.local.parking_local_general.domain.local.po.ThsChargingRulesPO;
import com.huaching.parking.local.parking_local_general.enums.local.RequestChargeTypeEnums;
import com.huaching.parking.local.parking_local_general.util.local.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.util.Comparator;
import java.util.Date;
import java.util.List;

/**
 * @program: parking_local_general
 * @description: 按照时段收费
 * @author: lidongsheng
 * @createData: 2020-04-26 10:38
 * @updateAuthor: lidongsheng
 * @updateData: 2020-04-26 10:38
 * @updateContent: 按照时段收费
 * @Version: 1.0.0
 * ************************************************
 * Copyright @ 李东升 2020. All rights reserved
 * ************************************************
 */
@Component
public class ThsTimeFramePayableMoney implements ThsRequestChargeAbstract {

    private static final Logger logger = LoggerFactory.getLogger(ThsTimeFramePayableMoney.class);

    @Override
    public Integer getPayableMoney(ThsChargingRulesPO thsChargingRulesPO, Date inTime) {
        String chargeDetails = thsChargingRulesPO.getChargeDetails();
        List<TimeFrameChargeDetail> list = JSON.parseArray(chargeDetails, TimeFrameChargeDetail.class);
        return timeFramePayableMoney(list, inTime, thsChargingRulesPO.getFreeTime()
                , thsChargingRulesPO.getMaxOneTimeMoney(), thsChargingRulesPO.getMaxOneDayMoney());
    }

    @Override
    public Integer getType() {
        return RequestChargeTypeEnums.TIME_FRAME.getCode();
    }

    /**
     * 按时段收费
     *
     * @param list            时段list
     * @param inTime          进场时间
     * @param freeTime        免费时长（分钟）
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @param maxOneDayMoney  一天停车的最高费用（单位分）
     * @return
     */
    public Integer timeFramePayableMoney(List<TimeFrameChargeDetail> list, Date inTime,
                                         Integer freeTime, Integer maxOneTimeMoney, Integer maxOneDayMoney) {
        list.sort(Comparator.comparing(o -> DateUtils.stringConvertDate("2020-01-01 " + o.getBeginTime(), 3)));
        inTime = DateUtils.getNextMinute(inTime, freeTime);
        //得到跨天的天数
        int acrossDayNum = (int) DateUtils.subDate(DateUtils.getNextDayStart(inTime, 0),
                DateUtils.getNextDayStart(new Date(), 0), 4);
        //是否存在跨天
        if (acrossDayNum > 0) {
            logger.info("跨天费用计算");
            return acrossDayTimeFrameCharge(list, inTime, acrossDayNum, maxOneTimeMoney, maxOneDayMoney);
        } else {
            logger.info("非跨天费用计算");
            return noAcrossDayTimeFrameCharge(list, inTime, new Date(), maxOneTimeMoney, maxOneDayMoney);
        }
    }

    /**
     * 按时段收费没有跨天的费用计算
     *
     * @param list
     * @param inTime          计费开始时间
     * @param endTime         计费结束时间
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @param maxOneDayMoney  一天停车的最高费用（单位分）
     * @return
     */
    private Integer noAcrossDayTimeFrameCharge(List<TimeFrameChargeDetail> list, Date inTime, Date endTime
            , Integer maxOneTimeMoney, Integer maxOneDayMoney) {
        int sumMoney = 0;

        String tempInTime = DateUtils.dateFormatToString(inTime, 3) + " ";
        String tempEndTime = DateUtils.dateFormatToString(endTime, 3) + " ";
        for (TimeFrameChargeDetail detail : list) {
            detail.beginTime = detail.beginTime + ":00";
            detail.endTime = detail.endTime + ":00";
            Date detailBeginTime = DateUtils.stringConvertDate(tempInTime + detail.beginTime, 3);
            Date detailEndTime = DateUtils.stringConvertDate(tempEndTime + detail.endTime, 3);

            if (inTime.compareTo(detailBeginTime) >= 0
                    && endTime.compareTo(detailEndTime) <= 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, inTime, endTime);
                break;
            }
            if (inTime.compareTo(detailBeginTime) >= 0
                    && inTime.compareTo(detailEndTime) < 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, inTime, detailEndTime);
                continue;
            }

            if (inTime.compareTo(detailBeginTime) < 0
                    && endTime.compareTo(detailEndTime) >= 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, null, null);
                continue;
            }
            if (endTime.compareTo(detailBeginTime) > 0
                    && endTime.compareTo(detailEndTime) < 0) {
                sumMoney += frameTypeTimeFrameCharge(detail, detailBeginTime, endTime);
                break;
            }

        }
        sumMoney = sumMoney > maxOneDayMoney ? maxOneDayMoney : sumMoney;
        sumMoney = sumMoney > maxOneTimeMoney ? maxOneTimeMoney : sumMoney;
        return sumMoney;
    }

    /**
     * 按时段收费跨天的费用计算
     *
     * @param list
     * @param inTime          开始时间（进场时间+freeTime）
     * @param acrossDayNum    跨天的天数
     * @param maxOneTimeMoney 一次停车的最高费用（单位分）
     * @param maxOneDayMoney  一天停车的最高费用（单位分）
     * @return
     */
    private Integer acrossDayTimeFrameCharge(List<TimeFrameChargeDetail> list, Date inTime, Integer acrossDayNum
            , Integer maxOneTimeMoney, Integer maxOneDayMoney) {

        int sumMoney = 0;
        //是否跨多天
        if (acrossDayNum > 1) {
            //跨天先计算出一整天需要缴费多少。
            int oneDayMoney = oneDayTimeFrameCharge(list);
            logger.info("停车完整一天收费：" + oneDayMoney);
            oneDayMoney = oneDayMoney > maxOneDayMoney ? maxOneDayMoney : oneDayMoney;
            //所跨天数的费用
            int acrossDayMoney = oneDayMoney * (acrossDayNum - 1);
            sumMoney += acrossDayMoney;
            if (sumMoney > maxOneTimeMoney) {
                return maxOneTimeMoney;
            }
            logger.info("所跨天数停车的停车收费：" + sumMoney);
        }
        Date nowDate = new Date();
        //记录最后一个区段的计费信息
        TimeFrameChargeDetail lastDetail = list.get(list.size() - 1);

        long firstDayMoney = 0, endDayMoney = 0;

        //第一次循环是为了计算停车当天的总费用
        String tempInTime = DateUtils.dateFormatToString(inTime, 3) + " ";
        String tempNowTime = DateUtils.dateFormatToString(nowDate, 3) + " ";
        for (TimeFrameChargeDetail detail : list) {
            detail.beginTime = detail.beginTime + ":00";
            detail.endTime = detail.endTime + ":00";
            Date detailBeginTime = DateUtils.stringConvertDate(tempInTime + detail.beginTime, 3);
            Date detailEndTime = DateUtils.stringConvertDate(tempInTime + detail.endTime, 3);
            //最初开始区段
            if (inTime.compareTo(detailBeginTime) >= 0 &&
                    inTime.compareTo(detailEndTime) < 0
            ) {
                Date tempEnd = DateUtils.stringConvertDate(DateUtils.dateFormatToString(inTime, 3) + " " + detail.endTime, 3);
                firstDayMoney += frameTypeTimeFrameCharge(detail, inTime, tempEnd);
            }
            //最初区段内
            if (inTime.compareTo(detailBeginTime) < 0
            ) {
                firstDayMoney += frameTypeTimeFrameCharge(detail, null, null);
            }
        }
        firstDayMoney = firstDayMoney > maxOneDayMoney ? maxOneDayMoney : firstDayMoney;
        logger.info("停车当天的停车收费：" + firstDayMoney);
        sumMoney += firstDayMoney;
        //第二次循环是为了计算停车结束当天的总费用
        for (TimeFrameChargeDetail detail : list) {
            detail.beginTime = detail.beginTime + ":00";
            detail.endTime = detail.endTime + ":00";
            Date detailNowBeginTime = DateUtils.stringConvertDate(tempNowTime + detail.beginTime, 3);
            Date detailNowEndTime = DateUtils.stringConvertDate(tempNowTime + detail.endTime, 3);
            //最终结束区段
            if (nowDate.compareTo(detailNowBeginTime) > 0 &&
                    nowDate.compareTo(detailNowEndTime) <= 0
            ) {
                Date tempBegin = DateUtils.stringConvertDate(tempNowTime + " " + detail.beginTime, 3);
                endDayMoney += frameTypeTimeFrameCharge(detail, tempBegin, nowDate);
            }
            //结束区段内
            if (nowDate.compareTo(detailNowBeginTime) > 0
                    && nowDate.compareTo(detailNowEndTime) > 0
            ) {
                endDayMoney += frameTypeTimeFrameCharge(detail, null, null);
            }

        }
        endDayMoney = endDayMoney > maxOneDayMoney ? maxOneDayMoney : endDayMoney;
        logger.info("停车结束当天的停车收费：" + endDayMoney);
        sumMoney += endDayMoney;
        if (sumMoney > maxOneTimeMoney) {
            return maxOneTimeMoney;
        }
        return sumMoney;
    }

    /**
     * 按时段收费全时段的总费用
     *
     * @param list
     * @return
     */
    public Integer oneDayTimeFrameCharge(List<TimeFrameChargeDetail> list) {
        int oneDayMoney = 0;
        for (TimeFrameChargeDetail detail : list) {
            detail.beginTime = detail.beginTime + ":00";
            detail.endTime = detail.endTime + ":00";
            oneDayMoney += frameTypeTimeFrameCharge(detail, null, null);
        }
        return oneDayMoney;
    }

    /**
     * 按时段收费根据开始时间和结束时间的总费用
     *
     * @param detail
     * @param beginTime 如果有开始时间，就以给定的开始时间计算
     * @param endTime   如果有结束时间，就以给定的结束时间计算
     * @return
     */
    public long frameTypeTimeFrameCharge(TimeFrameChargeDetail detail, Date beginTime, Date endTime) {
        if (beginTime == null) {
            beginTime = DateUtils.stringConvertDate("2020-01-01 " + detail.beginTime, 3);
        }
        if (endTime == null) {
            endTime = DateUtils.stringConvertDate("2020-01-01 " + detail.endTime, 3);
        }
        switch (detail.frameType) {
            //按照分钟收费
            case 0:
                long minute = DateUtils.subDate(beginTime, endTime, 2);
                return minute * detail.money;
            //按照小时收费
            case 1:
                long hour = DateUtils.subDate(beginTime, endTime, 3);
                return hour * detail.money;
            //按照固定收费
            case 2:
                return detail.money;
        }
        return 0;
    }

    /**
     * 按时段收费
     * "frameType": 0
     * "beginTime": "00:00",
     * "endTime": "07:00",
     * 表示在00：00-07：00时间端内每分钟收费2元
     * "money": 200
     * <p>
     * "frameType": 2
     * "beginTime": "00:00",
     * "endTime": "07:00",
     * 表示在00：00-07：00时间端内一共收费3元
     * "money": 300
     * <p>
     * "frameType": 0 按分钟收费
     * "frameType": 1 按小时收费
     * "frameType": 2 按固定收费
     */
    static class TimeFrameChargeDetail {

        private String beginTime;
        private String endTime;
        private Integer money;
        private Integer frameType;

        public Integer getFrameType() {
            return frameType;
        }

        public void setFrameType(Integer frameType) {
            this.frameType = frameType;
        }

        public String getBeginTime() {
            return beginTime;
        }

        public void setBeginTime(String beginTime) {
            this.beginTime = beginTime;
        }

        public String getEndTime() {
            return endTime;
        }

        public void setEndTime(String endTime) {
            this.endTime = endTime;
        }

        public Integer getMoney() {
            return money;
        }

        public void setMoney(Integer money) {
            this.money = money;
        }
    }
}
